﻿using System;
using System.Xml;
using System.Security.Cryptography;
using System.Security.Cryptography.Xml;
using System.Security.Cryptography.X509Certificates;
using System.Deployment.Internal.CodeSigning;

namespace IncomesRegisterSignatureSample
{   
    public class Signature
    {
        public Signature()
        {
            CryptoConfig.AddAlgorithm(typeof(RSAPKCS1SHA256SignatureDescription), SignedXml.XmlDsigRSASHA256Url);
        }
                
        public static void SignXmlDocument(XmlDocument documentToSign, string friendlyName)
        {
            var cert = GetCertificateFromCurrentUserStore(friendlyName);

            SignXmlDocument(documentToSign, cert, SignatureConfiguration.SHA256_RSASHA256_ExclusiveC14N_NoComments);
        }

        private static void SignXmlDocument(XmlDocument documentToSign, X509Certificate2 signingCertificate, SignatureConfiguration sigConfig)
        {
            if (documentToSign == null)
            {
                throw new ArgumentNullException(nameof(documentToSign));
            }
            if (documentToSign.DocumentElement == null)
            {
                throw new ArgumentException("Document element cannot be null", nameof(documentToSign));
            }
            if (signingCertificate == null)
            {
                throw new ArgumentNullException(nameof(signingCertificate));
            }
            if (sigConfig == null)
            {
                throw new ArgumentNullException(nameof(sigConfig));
            }

            var rsa = GetRsaForSigning(signingCertificate);

            var signedXml = new SignedXml(documentToSign)
            {
                SigningKey = rsa
            };

            // Define signature and canonicalization methods.
            signedXml.SignedInfo.SignatureMethod = sigConfig.SigningAlgorithmName;
            signedXml.SignedInfo.CanonicalizationMethod = sigConfig.CanonicalizationMethod;

            // Create reference and add set digest algorithm.
            var reference = new Reference(sigConfig.ReferenceUri)
            {
                DigestMethod = sigConfig.DigestAlgorithmName
            };

            // Add transforms for enveloped signature.
            reference.AddTransform(new XmlDsigEnvelopedSignatureTransform());                      
            signedXml.AddReference(reference);

            // Add keyinfo.
            var keyInfo = new KeyInfo();
            var keyInfoX509Data = new KeyInfoX509Data(signingCertificate);
            keyInfoX509Data.AddSubjectName(signingCertificate.SubjectName.Name);

            keyInfo.AddClause(keyInfoX509Data);

            signedXml.KeyInfo = keyInfo;

            // Compute signature and add to document.
            signedXml.ComputeSignature();
            var xmlDigitalSignature = signedXml.GetXml();
            documentToSign.DocumentElement.AppendChild(documentToSign.ImportNode(xmlDigitalSignature, true));
        }
                             
        /// <summary>
        /// Gets an RSA algorithm based on private key of the given certificate (to be used for signing). 
        /// </summary>
        private static RSA GetRsaForSigning(X509Certificate2 certificate)
        {
            if (certificate == null)
            {
                throw new ArgumentNullException(nameof(certificate));
            }
            if (!certificate.HasPrivateKey)
            {
                throw new ArgumentException("The given certificate must have a private key to be used for signing.", nameof(certificate));
            }

            return certificate.GetRSAPrivateKey();
        }

        private static X509Certificate2 GetCertificateFromX509Store(StoreName storeName, StoreLocation storeLocation, string friendlyName)
        {
            using (var store = new X509Store(storeName, storeLocation))
            {
                try
                {                   
                    store.Open(OpenFlags.ReadOnly);

                    var certs = store.Certificates.Find(X509FindType.FindByTimeValid, DateTime.Now, false);

                    if (certs.Count > 0)
                    {                        
                        foreach (var cert in certs)
                        {
                            if (String.CompareOrdinal(friendlyName, cert.FriendlyName) == 0)
                            {
                                return cert;
                            }
                        }
                    }                                        
                    return null;
                }
                finally
                {
                    store.Close();
                }
            }
        }

        public static X509Certificate2 GetCertificateFromCurrentUserStore(string subjectName)
        {
            return GetCertificateFromX509Store(StoreName.My, StoreLocation.CurrentUser, subjectName);
        }           
    }
}
